from contextlib import asynccontextmanager

import grpc
from loguru import logger


@asynccontextmanager
async def ErrorAsyncContextManager(context: any) -> any:
    try:
        yield None
    except Exception as e:
        message = type(e).__name__ + ": " + str(e)
        logger.error(message)
        await context.abort(grpc.StatusCode.UNKNOWN, message)


@asynccontextmanager
async def GrpcContextManager(grpc_stub, api_hostname: str, api_port: int = 8080) -> any:
    logger.debug(f"Connecting to grpc api: {api_hostname}:{api_port}")

    with grpc.insecure_channel(f"{api_hostname}:{api_port}") as channel:
        # create the client
        stub = grpc_stub(channel)

        # return the stub
        yield stub

    logger.debug(f"Disconnected from grpc api: {api_hostname}:{api_port}")


@asynccontextmanager
async def GrpcContextManagerAsync(
    grpc_stub, api_hostname: str, api_port: int = 8080
) -> any:
    logger.debug(f"Connecting to grpc api: {api_hostname}:{api_port}")

    async with grpc.aio.insecure_channel(f"{api_hostname}:{api_port}") as channel:
        # create the client
        stub = grpc_stub(channel)

        # return the stub
        yield stub

    logger.debug(f"Disconnected from grpc api: {api_hostname}:{api_port}")



from pydash.strings import snake_case


def dict_keys_to_snake_case(d):
    if isinstance(d, list):
        return [
            dict_keys_to_snake_case(i) if isinstance(i, (dict, list)) else i for i in d
        ]
    return {
        snake_case(a): dict_keys_to_snake_case(b) if isinstance(b, (dict, list)) else b
        for a, b in d.items()
    }



from beartype import beartype


@beartype
def poly_to_bbox(poly: list) -> dict:
    lats = [coord[0] for coord in poly]
    lons = [coord[1] for coord in poly]

    min_lat = min(lats)
    min_lon = min(lons)
    max_lat = max(lats)
    max_lon = max(lons)

    poly_bbox = {"top_left": (max_lat, min_lon), "bottom_right": (min_lat, max_lon)}

    return poly_bbox



from beartype import beartype


@beartype
def string_is_true_or_false(input_string: str) -> bool:
    ua = input_string.capitalize()
    if "True".startswith(ua):
        return True
    elif "False".startswith(ua):
        return False
    else:
        return False


@beartype
def string_to_bytes(input_str: str) -> bytes:
    return bytes(str(input_str).encode("utf-8"))


@beartype
def bytes_to_string(_bytes: bytes) -> str:
    return str(_bytes.decode("utf-8"))



import json
from loguru import logger


class BatchManager:
    def __init__(self, batch_size_kb=15000):
        # Set the batch size in KBs
        self.batch_size_kb = batch_size_kb
        # Calculate the batch size in bytes
        self.batch_size_bytes = batch_size_kb * 1024

        self.current_batch = []
        self.current_batch_size = 0

    def add_to_batch(self, document):
        # logger.debug("Adding document to batch")
        self.current_batch.append(document)
        self.current_batch_size += len(json.dumps(document).encode("utf-8"))

    def batch_has_docs(self):
        return len(self.current_batch) > 0

    def get_batch(self):
        return self.current_batch

    def reset_batch(self):
        self.current_batch = []
        self.current_batch_size = 0

    def is_document_too_large(self, document):
        doc_size_bytes = len(json.dumps(document).encode("utf-8"))

        if self.current_batch_size + doc_size_bytes > self.batch_size_bytes:
            logger.debug("Document exceeds batch size")
            return True
        else:
            # logger.debug("Document does not exceed batch size")
            return False



from beartype import beartype
from beartype.typing import Union

degrees_per_meters = 0.00006


@beartype
def meters_to_degrees(meters: Union[float, int]) -> Union[float, int]:
    return meters * degrees_per_meters



import json
from contextlib import asynccontextmanager

import asyncpg
from loguru import logger


@asynccontextmanager
async def DatabaseAsyncContextManager(db_pool: asyncpg.pool.Pool) -> asyncpg.Connection:
    # Perform setup that would go in __aenter__

    logger.debug("Acquiring a connection from the db_pool")
    async with db_pool.acquire() as connection:
        async with connection.transaction():
            logger.debug("Acquired a connection from the db_pool")
            # Set the type codec to use float instead of Decimal for numeric types
            await connection.set_type_codec(
                "numeric",
                encoder=str,
                decoder=float,
                schema="pg_catalog",
                format="text",
            )
            # Set the type codes to use instead of string for json types
            await connection.set_type_codec(
                "json", encoder=json.dumps, decoder=json.loads, schema="pg_catalog"
            )

            yield connection  # obj should be the object that will be bound in the as clause

    # Perform teardown that would go in __aexit__



from datetime import datetime

from ..time_helper import get_pb_ts_from_datetime


def dict_to_proto(dictionary, protobuf_message, field_types_mapping=None):
    if field_types_mapping is None:
        field_types_mapping = {}

    # Iterate over the key-value pairs in the dictionary
    for key, value in dictionary.items():
        # Check if the value is a datetime
        if isinstance(value, datetime):
            # Convert datetime to protobuf timestamp and set it using CopyFrom
            getattr(protobuf_message, key).CopyFrom(get_pb_ts_from_datetime(value))
        # Check if the value is a dictionary
        elif key in field_types_mapping and isinstance(value, dict):
            # If the key is in the mapping and the value is a dictionary, create a nested protobuf message of the specified type
            nested_message = field_types_mapping[key]
            getattr(protobuf_message, key).CopyFrom(
                dict_to_proto(value, nested_message, field_types_mapping)
            )
        else:
            # Set the field value in the Protobuf message
            setattr(protobuf_message, key, value)

    return protobuf_message
